/*
 * ============================================================================
 *  Title:    Debugger
 *  Author:   J. Zbiciak
 *  $Id: debug.c,v 1.18 1999/09/17 13:10:01 im14u2c Exp $
 * ============================================================================
 *  
 * ============================================================================
 *  This module will implement the debugger as a combination of interactive
 *  interface and a peripheral which maps into the entire address space
 *  which allows capturing memory events.
 * ============================================================================
 */

static const char rcs_id[]="$Id: debug.c,v 1.18 1999/09/17 13:10:01 im14u2c Exp $";

#include "config.h"
#include "periph/periph.h"
#include "cp1600/cp1600.h"
#include "cp1600/op_decode.h"
#include "debug.h"
#include "dasm1600.h"

static int dc_hits = 0, dc_miss = 0, dc_nocache = 0; 
static int dc_unhook_ok = 0, dc_unhook_odd = 0;

/*
 * ============================================================================
 *  DEBUG_RD         -- Capture/print a read event.
 *  DEBUG_WR         -- Capture/print a write event.
 *  DEBUG_TK         -- Debugger 'tick' function.  This where the user command
 *                      line called from.
 * ============================================================================
 */
uint_32 debug_rd(periph_t *p, periph_t *r, uint_32 a, uint_32 d)
{
    debug_t *debug = (debug_t*)p;

    if (debug->show_rd && p != r->req)
    {
        printf(" RD a=%.4X d=%.4X %-16s\n", 
                a, d, r == (periph_p)p->bus ? r->req->name : r->name);
    }

    return ~0U;
}

void    debug_wr(periph_t *p, periph_t *r, uint_32 a, uint_32 d)
{
    debug_t *debug = (debug_t*)p;

    if (debug->show_wr && p != r->req)
    {
        printf(" WR a=%.4X d=%.4X %-16s\n",
                a, d, r == (periph_p)p->bus ?  r->req->name : r->name);
    }
}

sint_32 debug_tk(periph_t *p, sint_32 len)
{
    debug_t *debug = (debug_t*)p;
    cp1600_t *cp = debug->cp1600;
    int     in_sync = len == 0;
    char    buf[1024], *s;
    int     i;
    static  uint_16 next_hex_dump = 0x0000;

    /* -------------------------------------------------------------------- */
    /*  See if we've caught up to the CPU, and if so, decrement our CPU     */
    /*  step count.                                                         */
    /* -------------------------------------------------------------------- */
    if (debug->tot_instr < cp->tot_instr)
    {
        in_sync = 1;
        if (debug->step_count > 0)
            debug->step_count--;

        debug->tot_instr = cp->tot_instr;
    }
    if (!len && debug->step_count)
        printf("Remaining step count: %d\n", debug->step_count);

    /* -------------------------------------------------------------------- */
    /*  If we're in-sync and the PC == a breakpoint address, drop into      */
    /*  debug mode.                                                         */
    /* -------------------------------------------------------------------- */
    if (in_sync && debug->bkpts)
    {
        for (i = 0; i < debug->bkpts; i++)
            if (debug->bkpt[i] == cp->r[7])
            {
                debug->step_count = 0;
                break;
            }
    }

    /* -------------------------------------------------------------------- */
    /*  If we're in sync with the CPU, then grab the current PC and words   */
    /*  at that location, disassemble and display.                          */
    /* -------------------------------------------------------------------- */
    if (in_sync && (debug->show_ins || !debug->step_count || !len))
    {
        uint_32 pc;
        char *dis;

show_disassem:
        pc = cp->r[7];

        /* ---------------------------------------------------------------- */
        /*  Where possible, cache disassembled instructions for speed.      */
        /* ---------------------------------------------------------------- */
        dis = debug_disasm(p, cp, pc, NULL, cp->D);
        dis = dis ? dis + 18 : "(null)";

        /* ---------------------------------------------------------------- */
        /*  Print the state of the machine, along w/ disassembly.           */
        /* ---------------------------------------------------------------- */
        printf(" %.4X %.4X %.4X %.4X %.4X %.4X %.4x %.4X %c%c%c%c%c%c%c%c"
               "%s\n",
               cp->r[0], cp->r[1], cp->r[2], cp->r[3], 
               cp->r[4], cp->r[5], cp->r[6], cp->r[7],
               cp->S ? 'S' : '-',
               cp->C ? 'C' : '-',
               cp->O ? 'O' : '-',
               cp->Z ? 'Z' : '-',
               cp->I ? 'I' : '-',
               cp->D ? 'D' : '-',
               cp->intr ? 'i' : '-',
               cp->irq  ? 'q' : '-', dis);
    }

    /* -------------------------------------------------------------------- */
    /*  If we've exhausted our CPU step count, drop into command prompt.    */
    /* -------------------------------------------------------------------- */
    if (debug->step_count <= 0 || len == 0)
    {
        int cmd = 0, c, arg = -1, arg2;
        extern void dump_state(void);

next_cmd:
        do
        {
            printf("> ");
            fflush(stdout);
            fgets(buf, 1023, stdin);
            s = buf;

            if (buf[0] == '\n')
            {
                cmd = arg = 1;
            }

            while (*s && isspace(*s)) s++;

            if (!*s) continue;

            c = toupper(*s);

            if (c == 'S') cmd = 1;  /* step */
            if (c == 'R') cmd = 2;  /* run  */
            if (c == 'D') cmd = 3;  /* dump/abort */
            if (c == 'Q') cmd = 4;  /* quit */
            if (c == 'B') cmd = 5;  /* breakpoint */
            if (c == 'M') cmd = 6;  /* mem dump */
            if (c == 'U') cmd = 7;  /* 'Un'assemble.            */
            if (c == 'C') cmd = 8;  /* Disassembly Cache stats. */
            if (c == 'G') cmd = 9;  /* Change re'G'ister value. */
            if (cmd < 3)
            {
                if (sscanf(s+1, "%d", &arg) != 1) 
                    arg=-1;
            } else if (cmd == 5) {
                if (sscanf(s+1, "%x", &arg) != 1) 
                    arg=-1;
            } else if (cmd == 6 || cmd == 7) {
                int args = sscanf(s+1, "%x %x", &arg, &arg2);

                if (args < 1) arg  = cmd == 6 ? next_hex_dump : cp->r[7];
                if (args < 2) arg2 = cmd == 6 ? 0x40          : 0x10;

                if (cmd == 6)
                    next_hex_dump = arg + arg2;
            } else if (cmd == 9) {
                int args = sscanf(s+1, "%x %x", &arg, &arg2);

                if (args != 2 || arg < 0 || arg > 7) arg = -1;
            } else 
                arg=0;
        } while (!cmd || arg < 0);

        switch (cmd)
        {   
            case 1: 
            case 2:
                debug->step_count = arg;
                debug->show_ins = debug->show_rd = debug->show_wr = cmd == 1;
                break;
            case 3: 
                dump_state();
                break;
            case 4:
                exit(0);
                break;
            case 5:
                if (debug->bkpts < DEBUG_MAX_BKPTS)
                {
                    printf("Set breakpoint at $%.4X\n", arg);
                    debug->bkpt[debug->bkpts++] = arg;
                }
                goto next_cmd;
                break;
            case 6:
                debug_dispmem(p, arg, arg2);
                goto next_cmd;
                break;
            case 7:
                debug_disasm_mem(p, cp, arg, arg2);
                goto next_cmd;
                break;
            case 8:
                printf("dc hits: %6d  misses: %6d  nocache: %6d  "
                       "unhook: %6d vs %6d\n", dc_hits, dc_miss, dc_nocache,
                       dc_unhook_ok, dc_unhook_odd);
                goto next_cmd;
                break;
            case 9:
                cp->r[arg] = arg2;
                goto show_disassem;
                break;
        }
    }

    return len;
}



/*
 * ============================================================================
 *  Instruction Disassembly Cache
 *
 *  This holds a cache of disassembled instructions.  The cache is maintained
 *  as an LRU list of entries.  We place an upper bound on the number of 
 *  entries just so we don't totally thrash through memory.
 * ============================================================================
 */
#define DISASM_CACHE (64)      /* Cache size. */

typedef struct disasm_cache_t
{
    char                    disasm[48];
    uint_32                 len;
    struct disasm_cache_t   *next,*prev;
    char                    **hook;
} disasm_cache_t;

disasm_cache_t *disasm_cache = NULL;


/*
 * ============================================================================
 *  DEBUG_DISASM      -- Disassembles one instruction, returning a pointer
 *                       to the disassembled text.  Uses the disassembly
 *                       cache if possible.
 * ============================================================================
 */
char * debug_disasm(periph_t *p, cp1600_t *cp, uint_16 addr, 
                    uint_32 *len, int dbd)
{
    static char buf[1024];
    uint_16 w1, w2, w3, pc = addr;
    disasm_cache_t *disasm = NULL;
    int instr_len;

#if 1
    /* -------------------------------------------------------------------- */
    /*  If we can't cache this disassembly, or if it's not disassembled     */
    /*  yet, go do the disassembly.  Also force disassembly if DBD is set   */
    /*  since we might have seen this instr before w/out DBD set.           */
    /* -------------------------------------------------------------------- */
    if (!cp->disasm[pc] || dbd)
    {
        w1 = periph_read((periph_t*)p->bus, p, pc    , ~0);
        w2 = periph_read((periph_t*)p->bus, p, pc + 1, ~0);
        w3 = periph_read((periph_t*)p->bus, p, pc + 2, ~0);

        instr_len = dasm1600(buf, pc, dbd, w1, w2, w3);

        dc_miss++;
        dc_hits--;            /* compensate for dc_hits++ in LRU update */

        /* ------------------------------------------------------------ */
        /*  If DBD is set, or if this instruction hasn't been decoded   */
		/*  yet, don't cache this instruction.                          */
        /* ------------------------------------------------------------ */
        if (dbd || !cp->instr[pc])
        {
            if (len) *len = instr_len;
            return buf + 17;
        }

        /* ------------------------------------------------------------ */
        /*  Allocate a new disassembly record by evicting the LRU.      */
        /* ------------------------------------------------------------ */
        disasm = disasm_cache->prev;        /* Tail of LRU list.        */

        if (disasm->hook &&                 /* Is this already hooked?  */
            *disasm->hook == (char*)disasm) /* Unhook it.               */
        {
            dc_unhook_ok++;
            *disasm->hook = NULL;
        } else
            dc_unhook_odd++;

        /* ------------------------------------------------------------ */
        /*  Hook this disassembly to the instruction record.            */
        /* ------------------------------------------------------------ */
        disasm->hook   = &cp->disasm[pc];
        cp->disasm[pc] = (char*) disasm;

        strncpy(disasm->disasm, buf + 17, sizeof(disasm->disasm));
        disasm->disasm[sizeof(disasm->disasm) - 1] = 0;
        disasm->len = instr_len;
    
    }

    /* -------------------------------------------------------------------- */
    /*  Grab the cached disassembly and update the LRU.                     */
    /* -------------------------------------------------------------------- */
    dc_hits++;
    disasm = (disasm_cache_t *)cp->disasm[pc];

    /* -------------------------------------------------------------------- */
    /*  Move this disasm record to head of the LRU.  Do this by first       */
    /*  unhooking it from the doubly-linked list, and then reinserting      */
    /*  it at the head of the doubly-linked list.                           */
    /* -------------------------------------------------------------------- */
    disasm->next->prev = disasm->prev;      /* Unhook 'next'.           */
    disasm->prev->next = disasm->next;      /* Unhook 'prev'.           */

    disasm->next = disasm_cache->next;      /* Hook us to new 'next'.   */
    disasm->prev = disasm_cache;            /* Hook us to new 'prev'.   */

    disasm->next->prev = disasm;            /* Hook new 'next' to us.   */
    disasm_cache->next = disasm;            /* Hook new 'prev' to us.   */


    /* -------------------------------------------------------------------- */
    /*  Returned cached disassembly.                                        */
    /* -------------------------------------------------------------------- */
    if (len) *len = disasm ? disasm->len : 0;
    return disasm ? disasm->disasm : NULL;
#else
    w1 = periph_read((periph_t*)p->bus, p, pc    , ~0);
    w2 = periph_read((periph_t*)p->bus, p, pc + 1, ~0);
    w3 = periph_read((periph_t*)p->bus, p, pc + 2, ~0);

    instr_len = dasm1600(buf, pc, cp->D, w1, w2, w3);

    if (len) *len = instr_len;

    return buf + 17;
#endif
}

/*
 * ============================================================================
 *  DEBUG_DISASM_MEM  -- Disassembles a range of memory locations.
 * ============================================================================
 */
void debug_disasm_mem(periph_t *p, cp1600_t *cp, uint_16 addr, uint_32 cnt)
{
    char * dis;
    uint_32 tot = 0, len = ~0, w0 = 0;

    while (tot <= cnt && len > 0)
    {
        if (len == 1)
            w0 = periph_read((periph_t*)p->bus, p, addr-1, ~0);

        dis = debug_disasm(p, cp, addr, &len, len == 1 && w0 == 0x0001);

        printf("    $%.4X:   %s\n", addr, dis);

        addr += len;
        tot  += len;
    }
}

/*
 * ============================================================================
 *  DEBUG_DISPMEM     -- Displays ten lines of "hex dump" memory information.
 *                       The first arg is the address to start dumping at.
 *                       The second arg is the number of addresses to dump.
 * ============================================================================
 */
void debug_dispmem(periph_t *p, uint_16 addr, uint_16 len)
{
    int         i, j, k, l;
    uint_32     w[8];
    
    /* -------------------------------------------------------------------- */
    /*  Round our address down to the next lower multiple of 8.  Add this   */
    /*  difference to the length, and then round the length up to the       */
    /*  next multiple of 8.                                                 */
    /* -------------------------------------------------------------------- */
    k = addr & ~7;

    len = (len + 7 + addr - k) & ~7;
    
    /* -------------------------------------------------------------------- */
    /*  Iterate until we've dumped the entire block of memory.              */
    /* -------------------------------------------------------------------- */
    for (i = 0; i < len; i += 8)
    {
        l = k;
        
        /* ---------------------------------------------------------------- */
        /*  Read one row of 8 memory locations.                             */
        /* ---------------------------------------------------------------- */
        for (j = 0; j < 8; j++)
        {
            w[j] = periph_read((periph_t*)p->bus, p, k++, ~0);
        }
        
        /* ---------------------------------------------------------------- */
        /*  Display them.                                                   */
        /* ---------------------------------------------------------------- */
        printf( "%.4X:  %.4X %.4X %.4X %.4X  %.4X %.4X %.4X %.4X   # ",
                    l, w[0], w[1], w[2], w[3], w[4], w[5], w[6], w[7] );
        
        /* ---------------------------------------------------------------- */
        /*  Display ASCII equivalents.                                      */
        /* ---------------------------------------------------------------- */
        for (j = 0; j < 8; j++)
        {
            char c, d;
        
            c = w[j] >> 8;
            d = w[j];
            
            if ( !isalnum(c) )
                c = '.';
            if ( !isalnum(d) )
                d = '.';
                
            printf( "%c%c", c, d );
        }
        
        printf( "\n" );
    }
}


/*
 * ============================================================================
 *  DEBUG_INIT       -- Initializes a debugger object and registers a CPU
 *                      pointer.
 * ============================================================================
 */
int     debug_init(debug_t *debug, cp1600_t *cp1600)
{
    /* -------------------------------------------------------------------- */
    /*  Set up the debugger's state.                                        */
    /* -------------------------------------------------------------------- */
    debug->periph.read     = debug_rd;
    debug->periph.write    = debug_wr;
    debug->periph.tick     = debug_tk;
    debug->periph.min_tick = 1;
    debug->periph.max_tick = 4;

    debug->periph.addr_base = 0;
    debug->periph.addr_mask = ~0U;

    debug->show_rd = 1;
    debug->show_wr = 1;
    debug->cp1600  = cp1600;

    /* -------------------------------------------------------------------- */
    /*  Set up the instruction disassembly cache.                           */
    /* -------------------------------------------------------------------- */
    if (!disasm_cache)
    {
        int i;

        disasm_cache = calloc(sizeof(disasm_cache_t), DISASM_CACHE);
        
        for (i = 0; i < DISASM_CACHE-1; i++)
        {
            disasm_cache[i  ].next = &disasm_cache[i+1];
            disasm_cache[i+1].prev = &disasm_cache[i  ];
        }
        disasm_cache[0].prev = &disasm_cache[DISASM_CACHE-1];
        disasm_cache[DISASM_CACHE-1].next = &disasm_cache[0];
    }

    return 0;
}
/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */
