/*
 * ============================================================================
 *  Title:    STIC -- Standard Television Interface Chip, AY-3-8900
 *  Author:   J. Zbiciak
 *  $Id: stic.c,v 1.14 1999/09/13 02:35:10 im14u2c Exp $
 * ============================================================================
 *  This module implements the STIC Chip.  Its functionality is currently
 *  far from complete.  For instance, I don't implement MOBs yet.
 * ============================================================================
 *  The STIC processor is active over several regions of memory:
 *
 *   0x0000 - 0x001F     -- MOB (Movable OBject) controls 
 *                          0x0000 + n  X Position, MOB #n
 *                          0x0008 + n  Y Position, MOB #n
 *                          0x0010 + n  Attribute Register, MOB #n
 *                          0x0018 + n  Collision Detect, MOB #n
 *
 *   0x0020 - 0x0021     -- STIC control registers  
 *                          0x0020      Display Enable
 *                          0x0021      Graphics Mode
 *
 *   0x0028 - 0x002C     -- Global color settings
 *                          0x0028 + n  Color Stack (n == 0..3)
 *                          0x002C      Border Color
 *
 *   0x0030 - 0x0032     -- Display framing
 *                          0x0030      Pixel Column Delay (0..7)
 *                          0x0031      Pixel Row Delay (0..7)
 *                          0x0032      Edge inhibit (bit0->left, bit1->top)
 *
 *   0x0200 - 0x035F     -- Background Card Table (BACKTAB)
 *
 *   0x3800 - 0x3FFF     -- Graphics RAM  (0x200 entries, w/ 4 aliases.)
 *  
 *                          Note:  BACKTAB and GRAM are actually physically
 *                          separate.  This STIC implementation 'snoops'
 *                          these accesses for performance reasons.
 *
 *  For now, I handle all of the control register accesses into one 
 *  peripheral entry and all bus-snooping accesses in a second peripheral
 *  entry.
 * ============================================================================
 */

static const char rcs_id[]="$Id: stic.c,v 1.14 1999/09/13 02:35:10 im14u2c Exp $";

#include "config.h"
#include "macros.h"
#include "periph/periph.h"
#include "mem/mem.h"
#include "cp1600/cp1600.h"
#include "debug/debug.h"
#include "gfx/gfx.h"

#include "stic.h"

#ifdef macintosh
/* Prototypes required on macintosh */
void efix(uint_16 *ary, int size);
#endif

/*
 * ============================================================================
 *  TEMPORARY CRUFT
 * ============================================================================
 */
int sprite_hack = 1;
static void stic_display    (stic_t *stic);
static void stic_temp_fgbg  (stic_t *stic);
static void stic_temp_colstk(stic_t *stic);

#define STIC_X (160)
#define STIC_Y (200)

/*
uint_32 mem_rd_null (periph_t *per, periph_t *ign, uint_32 addr, uint_32 data);
void    mem_wr_null (periph_t *per, periph_t *ign, uint_32 addr, uint_32 data);
 */

/*
 * ============================================================================
 *  STIC_CTRL_RD     -- Read from a STIC control register.
 *  STIC_CTRL_WR     -- Write to a STIC control register.
 * ============================================================================
 */

uint_32 stic_ctrl_rd(periph_t *per, periph_t *ign, uint_32 addr, uint_32 data)
{
    stic_t *stic = (stic_t*)per->parent;

    (void)ign; 
    (void)data; 

    /* -------------------------------------------------------------------- */
    /*  For now, do nothing special -- just return the raw, encoded data.   */
    /* -------------------------------------------------------------------- */
    if (addr == 0x0021) 
    {
        /*printf("--> Setting Col Stk mode, RD 0x0021\n");*/
        if (stic->mode != 0) stic->bt_dirty = 1;
        stic->mode = 0;
        stic->upd  = stic_temp_colstk;
    }

    return addr >= 0x18 && addr <= 0x1F ? 0 : stic->cr.raw[addr];
}

void    stic_ctrl_wr(periph_t *per, periph_t *ign, uint_32 addr, uint_32 data)
{
    stic_t *stic = (stic_t*)per->parent;
    uint_32 old;

    (void)ign; 

    /* -------------------------------------------------------------------- */
    /*  Capture the control-register write and store it in its raw,         */
    /*  encoded form.                                                       */
    /* -------------------------------------------------------------------- */
    old = stic->cr.raw[addr];
    stic->cr.raw[addr] = data;

    /* -------------------------------------------------------------------- */
    /*  At this point, we might do something else depending on what CR      */
    /*  this was.  For now, the only thing I have implemented is the        */
    /*  Vid-Enable flag stuff.                                              */
    /* -------------------------------------------------------------------- */
    if (addr == 0x0020)
        stic->ve_post = 1;
    else if (addr == 0x0021) 
    {
        if (stic->mode == 0) stic->bt_dirty = 1;
        /*printf("--> Setting FG / BG mode, WR 0x0021 <- 0x%.4X\n", data);*/
        stic->mode = 1;
        stic->upd  = stic_temp_fgbg;
    } else if (0xFFFF & (old-data))
        stic->bt_dirty = 1;

    return;
}


/*
 * ============================================================================
 *  STIC_BTAB_WR         -- Monitor writes to BACKTAB
 *  STIC_GRAM_WR         -- Monitor writes to GRAM
 * ============================================================================
 */


void    stic_btab_wr(periph_t *per, periph_t *ign, uint_32 addr, uint_32 data)
{
    stic_t *stic = (stic_t*)per->parent;

    /* -------------------------------------------------------------------- */
    /*  TODO:  Right now, I'm just using this as little more than a stub.   */
    /* -------------------------------------------------------------------- */
    (void)per; 
    (void)ign; 

    if (addr < 240)
    {
        stic->backtab[addr] = data;
        stic->bt_dirty = 1;
    }
}

void    stic_gram_wr(periph_t *per, periph_t *ign, uint_32 addr, uint_32 data)
{
    stic_t *stic = (stic_t*)per->parent;
    /* -------------------------------------------------------------------- */
    /*  TODO:  Right now, I'm just using this as a stub.                    */
    /* -------------------------------------------------------------------- */
    (void)per; 
    (void)ign; 
    (void)addr; 
    (void)data; 
    stic->gr_dirty = 1;
}


/*
 * ============================================================================
 *  STIC_INIT        -- Initialize the STIC
 * ============================================================================
 */

int stic_init
(
    stic_t          *stic,
    mem_t           *grom,
    mem_t           *gram,
    int             *irq,
    gfx_t           *gfx
)
{
    /* -------------------------------------------------------------------- */
    /*  First, zero out the STIC structure to get rid of anything that      */
    /*  might be dangling.                                                  */
    /* -------------------------------------------------------------------- */
    memset((void*)stic, 0, sizeof(stic_t));

    /* -------------------------------------------------------------------- */
    /*  Set our graphics subsystem pointers.                                */
    /* -------------------------------------------------------------------- */
    stic->gfx  = gfx;
    stic->disp = gfx->vid;

    /* -------------------------------------------------------------------- */
    /*  Allocate memory for our BACKTAB shadow.                             */
    /* -------------------------------------------------------------------- */
    stic->backtab = calloc(2, 240);

    if (!stic->backtab)
        return -1;


    /* -------------------------------------------------------------------- */
    /*  Set up our internal flags.                                          */
    /* -------------------------------------------------------------------- */
    stic->phase = 0;
    stic->upd   = stic_temp_colstk;
    
    /* -------------------------------------------------------------------- */
    /*  Now, borrow the pointers into GRAM, GROM.  The STIC implementation  */
    /*  is a little more tightly coupled to the mem_t implementation than   */
    /*  it probably should be, but this is for efficiency sake.   I will    */
    /*  probably eventually embed the GROM and GRAM periph structures in    */
    /*  the stic_t itself.                                                  */
    /* -------------------------------------------------------------------- */
    stic->grom = grom->image;
    stic->gram = gram->image;

    /* -------------------------------------------------------------------- */
    /*  Record our IRQ output pointer.  This is usually &cp1600->irq.       */
    /* -------------------------------------------------------------------- */
    stic->irq  = irq;

    /* -------------------------------------------------------------------- */
    /*  Now, set up our peripheral functions for the main STIC peripheral.  */
    /* -------------------------------------------------------------------- */
    stic->stic_cr.read      = stic_ctrl_rd;
    stic->stic_cr.write     = stic_ctrl_wr;
    stic->stic_cr.tick      = stic_tick;
    stic->stic_cr.min_tick  = (STIC_FRAMCLKS + 8) / 16;
    stic->stic_cr.max_tick  = (STIC_FRAMCLKS + 8) / 16;
    stic->stic_cr.next_tick = 1;
    stic->stic_cr.addr_base = 0x00000000;
    stic->stic_cr.addr_mask = 0x0000003F;
    stic->stic_cr.parent    = (void*) stic;
    
    /* -------------------------------------------------------------------- */
    /*  Lastly, set up the 'snooping' STIC peripherals.                     */
    /* -------------------------------------------------------------------- */
    stic->snoop_btab.read       = NULL;
    stic->snoop_btab.write      = stic_btab_wr;
    stic->snoop_btab.tick       = NULL;
    stic->snoop_btab.min_tick   = ~0U;
    stic->snoop_btab.max_tick   = ~0U;
    stic->snoop_btab.addr_base  = 0x00000200;
    stic->snoop_btab.addr_mask  = 0x000000FF;
    stic->snoop_btab.parent     = (void*) stic;

    stic->snoop_gram.read       = NULL;
    stic->snoop_gram.write      = stic_gram_wr;
    stic->snoop_gram.tick       = NULL;
    stic->snoop_gram.min_tick   = ~0U;
    stic->snoop_gram.max_tick   = ~0U;
    stic->snoop_gram.addr_base  = 0x00003800;
    stic->snoop_gram.addr_mask  = 0x000001FF;
    stic->snoop_gram.parent     = (void*) stic;

    return 0;
}

/*
 * ============================================================================
 *  STIC_TICK        -- Perform a STIC update
 * ============================================================================
 */
sint_32 stic_tick
(
    periph_p        per,
    sint_32         len
)
{
    stic_t *stic = (stic_t*)per->parent;

    /* -------------------------------------------------------------------- */
    /*  Check which tick phase we're in.  Phase is determined by the value  */
    /*  of stic->phase at the start of a tick.                              */
    /*   -- Phase 0 is Start of VBlank.  (End of Active Display.)           */
    /*   -- Phase 4 is Start of Active Display.  (End of VBlank.)           */
    /*  Since phases are toggled inside the tick function after sensing     */
    /*  the current phase, between ticks stic->phase has a slightly         */
    /*  different meaning:                                                  */
    /*   -- Phase 1..4 is VBlank In Progress.                               */
    /*   -- Phase 5..0 is Active Display In Progress.                       */
    /* -------------------------------------------------------------------- */ 
    if (stic->phase == 0)
    {
#if 0
        /* ---------------------------------------------------------------- */ 
        /*  Debugging -- report how long vid_enable was false.              */
        /* ---------------------------------------------------------------- */ 
        {
            static int disabled = 0;
            if (!stic->vid_enable) 
                disabled++; 
            else if (disabled)
            {
                printf("Screen was disabled for %d frames\n", disabled);
                disabled = 0;
            }
        }
#endif

        /* ---------------------------------------------------------------- */ 
        /*  Do VBlank things:                                               */
        /*   -- Render and show the active display.                         */
        /*   -- Kick the IRQ line.                                          */
        /*   -- Clear the video enable.                                     */
        /*   -- Clear the dirty bits.                                       */
        /* ---------------------------------------------------------------- */ 
        if (stic->upd) stic->upd(stic);
        if (stic->irq) *stic->irq = 1;

        /*stic->vid_enable      = 0;*/
        stic->bt_dirty          = 0;
        stic->gr_dirty          = 0;
        stic->ob_dirty          = 0;

    } else if (stic->phase == 5)
    {
        /* ---------------------------------------------------------------- */ 
        /*  Do Active Display things:                                       */
        /*   -- Deassert the IRQ line.                                      */
        /*   -- TEMP HACK: If we've changed vid-enable states, we're dirty  */
        /*   -- Snapshot the video enable for the next frame.               */
        /* ---------------------------------------------------------------- */ 

        if (stic->irq)  *stic->irq = 0;

        if (stic->vid_enable != stic->ve_post)
            stic->bt_dirty = 1;

        stic->vid_enable        = stic->ve_post;
        stic->ve_post           = 0;
    } 

    /* -------------------------------------------------------------------- */ 
    /*  Toggle the phases and get out of here.                              */
    /* -------------------------------------------------------------------- */ 
    stic->phase = (stic->phase + 1) & 15;

    return len;
}

/*
 * ============================================================================
 *  STIC_TEMP_FGBG   -- Temporary implementation of FG/BG mode
 * ============================================================================
 */
static void stic_temp_fgbg(stic_t *stic)
{
    int i, j;
    uint_32 x, y, xx, yy, pat, fg, bg, card; 
    uint_16 *grfx;

    /* -------------------------------------------------------------------- */
    /*  For now, brute-force render the display.                            */
    /* -------------------------------------------------------------------- */
    if (!(stic->bt_dirty+stic->gr_dirty))
        return;


    /* -------------------------------------------------------------------- */
    /*  Now draw the 'cards'                                                */
    /* -------------------------------------------------------------------- */
    for (y = 0; y < 12; y++)
    {
        for (x = 0; x < 20; x++)
        {
            uint_8 *row0, *row1;

            card = stic->backtab[x + y*20] & 0x3FFF;
            pat = (card >> 3) & 0x1FF;

            pat &= 0x13F;
            if (pat & 0x100) 
            { 
                grfx = stic->gram; pat -= 0x100; 
            } else 
                grfx = stic->grom;
            pat <<= 3;

            fg = card & 0x7;

            bg = ((card >> 9) & 0xB) | ((card >> 11) & 0x4);

            row0 = stic->disp + ((y * STIC_X)<<4) + 4 * STIC_X;
            row1 = row0 + STIC_X;
            for (yy = (y << 4), j = 0; j < 8; yy+=2, j++)
            {
                for (xx = (x << 3), i = 0; i < 8; xx++, i++)
                {
                    row0[xx] = row1[xx] = 
                        (0x80 & (grfx[pat + j] << i)) ? fg : bg;
                }
                row0 += STIC_X * 2;
                row1 += STIC_X * 2;
            }
        }
    }

    stic_display(stic);
    return;
}

/*
 * ============================================================================
 *  STIC_TEMP_COLSTK -- Temporary implementation of Color Stack
 * ============================================================================
 */
static void stic_temp_colstk(stic_t *stic)
{
    int i, j, csq = 0;
    uint_32 x, y, xx, yy, pat, fg, bg, col_stack = 0, card; 
    uint_32 cs0=0, cs1=0, cs2=0, cs3=0;
    uint_16 *grfx;

    /* -------------------------------------------------------------------- */
    /*  For now, brute-force render the display.                            */
    /* -------------------------------------------------------------------- */

    if (!(stic->bt_dirty+stic->gr_dirty))
        return;

#if 0
    /* -------------------------------------------------------------------- */
    /*  Set the background color.                                           */
    /* -------------------------------------------------------------------- */
    memset(stic->disp, bg = stic->cr.raw[0x2C], STIC_X * STIC_Y);
#endif


#if 0
    if (!vid_enable)
    {
        memset(stic->disp, bg = stic->cr.raw[0x2C], STIC_X * STIC_Y);
        return;
    }
#endif

    /* -------------------------------------------------------------------- */
    /*  Now draw the 'cards'                                                */
    /* -------------------------------------------------------------------- */
    col_stack = 0;
    for (y = 0; y < 12; y++)
    {
        for (x = 0; x < 20; x++)
        {
            csq = 0;
            card = stic->backtab[x + y*20] & 0x3FFF;
            pat = (card >> 3) & 0x1FF;

            if (pat > 0xFF) pat &= 0x13F;
            if (pat & 0x100) 
            { 
                grfx = stic->gram; pat -= 0x100; 
            } else 
                grfx = stic->grom;
            pat <<= 3;

            fg = card & 0x7;


            csq = ((card >> 11) & 3) == 2;
            if (csq)
            {
                uint_8 *row0, *row1;

                row0 = stic->disp + ((y * STIC_X)<<4) + 4 * STIC_X;
                row1 = row0 + STIC_X * 8;

                cs0 = card & 7;
                cs1 = (card >> 3) & 7;
                cs2 = (card >> 6) & 7;
                cs3 = ((card >> 9) & 3) | ((card >> 11) & 4);
                if (cs0 == 7) cs0 = stic->cr.raw[0x28 + (col_stack&3)] & 15;
                if (cs1 == 7) cs1 = stic->cr.raw[0x28 + (col_stack&3)] & 15;
                if (cs2 == 7) cs2 = stic->cr.raw[0x28 + (col_stack&3)] & 15;
                if (cs3 == 7) cs3 = stic->cr.raw[0x28 + (col_stack&3)] & 15;

                for (yy = (y << 4), j = 0; j < 8; yy++, j++)
                {
                    for (xx = (x << 3), i = 0; i < 4; xx++, i++)
                    {
                        row0[xx    ] = cs0; row0[xx + 4] = cs1;
                        row1[xx    ] = cs2; row1[xx + 4] = cs3;
                    }
                    row0 += STIC_X;
                    row1 += STIC_X;
                }
            } else
            {
                uint_8 *row0, *row1;

                col_stack += (card >> 13) & 1;
                fg |= (card >> 9) & 0x8;
                bg = stic->cr.raw[0x28 + (col_stack & 3)] & 15;

                row0 = stic->disp + ((y * STIC_X)<<4) + 4 * STIC_X;
                row1 = row0 + STIC_X;
                for (yy = (y << 4), j = 0; j < 8; yy+=2, j++)
                {
                    for (xx = (x << 3), i = 0; i < 8; xx++, i++)
                    {
                        row0[xx] = row1[xx] = 
                            (0x80 & (grfx[pat + j] << i)) ? fg : bg;
                    }
                    row0 += STIC_X * 2;
                    row1 += STIC_X * 2;
                }
            }
        }
    }

    stic_display(stic);
    return ;
}



/*
 * ============================================================================
 *  STIC_DISPLAY     -- Final fixups on the display before we're done.  
 * ============================================================================
 */
static void stic_display(stic_t *stic)
{
    int x, y;
    int x_dly = stic->cr.raw[0x30] & 7;
    int y_dly =(stic->cr.raw[0x31] & 7) * 2;
    int bord  = stic->cr.raw[0x2C] & 15;
    int block = stic->cr.raw[0x32] & 3;
    static int last_bord = -1;
    static int last_enable = -1;


    /* This function is totally lame. Exceedingly lame. But it runs. Barely. */

    stic->gfx->dirty = 1;

    if (last_enable != stic->vid_enable)
        gfx_vid_enable(stic->gfx, stic->vid_enable);

    last_enable = stic->vid_enable;

    gfx_set_bord(stic->gfx, block & 1, block & 2, x_dly, y_dly, bord);

    /* Sprite Rect Hack: THIS IS VERY TEMPORARY AND IS SLOWS THE EMU ALOT */
    if (1)
    {
        int i, x0, y0, x1, y1;
        uint_8 *pix0, *pix1;

        for (i = 0; i < 8; i++)
        {
            int yh;

            stic->cr.raw[0x18 + i] &= ~0x100; 
            if ((stic->cr.raw[i] & 0x200) == 0)
                continue;

            x0 = (stic->cr.raw[i] & 0xFF) - 8 + x_dly;
            x1 = (stic->cr.raw[i] & 0xFF) + x_dly + 
                    (stic->cr.raw[i] & 0x400 ? 8 : 0);

            yh = (stic->cr.raw[i+8] & 0x80) ? 8 : 4;
            yh <<= (stic->cr.raw[i+8] & 0x300) >> 8;

            y0 = (stic->cr.raw[i+8] & 0x7F) - 8 + y_dly;
            y1 = (stic->cr.raw[i+8] & 0x7F) - 8 + y_dly + yh;

            if (x0 <= 0 || y0 <= 0 || x1 >= 159 || y1 >= 95)
                stic->cr.raw[0x18 + i] |=  0x100; 

            if (sprite_hack)
            {
                uint_32 c = stic->cr.raw[i+0x10];

                c = (c & 7) | ((c >> 9) & 8);

                if (x0 <   0) x0 =   0;
                if (x1 <   0) x1 =   0;
                if (y0 <   0) y0 =   0;
                if (y1 <   0) y1 =   0;
                if (x0 > 159) x0 = 159;
                if (x1 > 159) x1 = 159;
                if (y0 >  95) y0 =  95;
                if (y1 >  95) y1 =  95;

                pix0 = stic->disp + y0 * STIC_X * 2 + STIC_X * 4;
                pix1 = stic->disp + y1 * STIC_X * 2 + STIC_X * 4;
                for (x = x0; x <= x1; x += 2)
                    pix0[x] = pix1[x] = 16;
                for (x = x0+1; x <= x1-1; x += 2)
                    pix0[x] = pix1[x] = c;

                pix0 = stic->disp + y0 * STIC_X * 2 + x0 + STIC_X * 4;
                pix1 = stic->disp + y0 * STIC_X * 2 + x1 + STIC_X * 4;
                if (y1 > y0) for (y = y1 - y0 ; y >= 0; y--)
                {
                    *pix0 = *pix1 = 16;
                    pix0 += STIC_X * 2;
                    pix1 += STIC_X * 2;
                }
                pix0 = stic->disp + y0 * STIC_X * 2 + x0 + STIC_X * 5;
                pix1 = stic->disp + y0 * STIC_X * 2 + x1 + STIC_X * 5;
                if (y1-1 > y0) for (y = y1 - y0 - 1 ; y >= 0; y--)
                {
                    *pix0 = *pix1 = c;
                    pix0 += STIC_X * 2;
                    pix1 += STIC_X * 2;
                }
            }
        }
    }

    if (last_bord != bord)
    {
        memset(stic->disp,                bord, STIC_X * 4);
        memset(stic->disp + STIC_X * 196, bord, STIC_X * 4);
    }

    last_bord = bord;
}


void efix(uint_16 *ary, int size)
{
#ifdef _BIG_ENDIAN
    (void)ary;
    (void)size;
    return;
#else
    int i;

    for (i = 0; i < size; i++)
        ary[i] = ((0xFF00 & ary[i]) >> 8) | ((0x00FF & ary[i]) << 8);
#endif
}


/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */
