/*
 * ============================================================================
 *  Title:    STIC -- Standard Television Interface Chip, AY-3-8900
 *  Author:   J. Zbiciak
 *  $Id: stic.h,v 1.9 1999/09/13 02:35:10 im14u2c Exp $
 * ============================================================================
 *  This module implements the STIC Chip.  Its functionality is currently
 *  far from complete.  For instance, I don't implement MOBs yet.
 * ============================================================================
 *  The STIC processor is active over several regions of memory:
 *
 *   0x0000 - 0x001F     -- MOB (Movable OBject) controls 
 *                          0x0000 + n  X Position, MOB #n
 *                          0x0008 + n  Y Position, MOB #n
 *                          0x0010 + n  Attribute Register, MOB #n
 *                          0x0018 + n  Collision Detect, MOB #n
 *
 *   0x0020 - 0x0021     -- STIC control registers  
 *                          0x0020      Display Enable
 *                          0x0021      Graphics Mode
 *
 *   0x0028 - 0x002C     -- Global color settings
 *                          0x0028 + n  Color Stack (n == 0..3)
 *                          0x002C      Border Color
 *
 *   0x0030 - 0x0032     -- Display framing
 *                          0x0030      Pixel Column Delay (0..7)
 *                          0x0031      Pixel Row Delay (0..7)
 *                          0x0032      Edge inhibit (bit0->left, bit1->top)
 *
 *   0x0200 - 0x035F     -- Background Card Table (BACKTAB)
 *
 *   0x3800 - 0x39FF     -- Graphics RAM
 *  
 *                          Note:  BACKTAB and GRAM are actually physically
 *                          separate.  This STIC implementation 'snoops'
 *                          these accesses for performance reasons.
 *
 *  For now, I handle all of the control register accesses into one 
 *  peripheral entry and all bus-snooping accesses in a second peripheral
 *  entry.
 * ============================================================================
 */

#ifndef _STIC_H
#define _STIC_H 1

/*
 * ============================================================================
 *  STIC_T           -- Main STIC structure.
 *
 *  Note: These bitfields seem to be broken at the present.
 *
 *  STIC_MOB_X_T     -- MOB X Position Bitfield Structure
 *  STIC_MOB_Y_T     -- MOB Y Position Bitfield Structure
 *  STIC_MOB_A_T     -- MOB Attribute Bitfield Structure
 *  STIC_MOB_C_T     -- MOB Collision Bitfield Structure
 * ============================================================================
 */

typedef struct stic_mob_x_t
{
    BFE(unsigned    x_pos       : 8,        /*  X Position                  */
    BFE(unsigned    unused0     : 1,        /*  Unused                      */
    BFE(unsigned    vis         : 1,        /*  Visibility                  */
    BFE(unsigned    x_size      : 1,        /*  X Size                      */
        unsigned    unused1     : 22))));   /*  Unused                      */
} stic_mob_x_t;

typedef struct stic_mob_y_t
{
    BFE(unsigned    y_pos       : 7,        /*  Y Position                  */
    BFE(unsigned    y_res       : 1,        /*  Y Resolution (8/16 rows)    */
    BFE(unsigned    y_size0     : 1,        /*  2x scanline height.         */
    BFE(unsigned    y_size1     : 1,        /*  4x scanline height.         */
    BFE(unsigned    x_flip      : 1,        /*  X-Axis flip                 */
    BFE(unsigned    y_flip      : 1,        /*  Y-Axis flip                 */
        unsigned    unused      : 21)))))); /*  Unused                      */
} stic_mob_y_t;

typedef struct stic_mob_a_t
{
    BFE(unsigned    fg210       : 3,        /*  Foreground Color lsbs       */
    BFE(unsigned    card        : 6,        /*  Card # (0..63)              */
    BFE(unsigned    unused0     : 2,        /*  Unused                      */
    BFE(unsigned    gram        : 1,        /*  0 == GROM, 1 == GRAM card   */
    BFE(unsigned    fg3         : 1,        /*  Foreground Color msb        */
    BFE(unsigned    prio        : 1,        /*  Priority (1=behind bg.)     */
        unsigned    unused1     : 18)))))); /*  Unused                      */
} stic_mob_a_t;

typedef struct stic_mob_c_t
{
    BFE(unsigned    mobs        : 8,        /*  Collision w/ mob 0..7       */
    BFE(unsigned    bg          : 1,        /*  Collision w/ background     */
    BFE(unsigned    bord        : 1,        /*  Collision w/ border         */
        unsigned    unused      : 22)));    /*  Unused                      */
} stic_mob_c_t;

typedef struct stic_t
{
    /* -------------------------------------------------------------------- */
    /*  The STIC is split into three separate peripherals, since a RD/WR    */
    /*  to something the STIC is interested in needs to go one of three     */
    /*  places -- either (1) to update the general state of the STIC via a  */
    /*  control register, (2) to update the BACKTAB, or (3) to update a     */
    /*  card definition in GRAM.                                            */
    /* -------------------------------------------------------------------- */
    periph_t    stic_cr;
    periph_t    snoop_btab;
    periph_t    snoop_gram;

    /* -------------------------------------------------------------------- */
    /*  The STIC registers themselves are decoded using an overlay between  */
    /*  the RAW registers and some bitfield structures, just as in the      */
    /*  CP-1600 'op_decode' routines.                                       */
    /* -------------------------------------------------------------------- */
    union
    {
        uint_32 raw[0x3F];
        struct
        {
            stic_mob_x_t    mob_x[8];
            stic_mob_y_t    mob_y[8];
            stic_mob_a_t    mob_a[8];
            stic_mob_c_t    mob_c[8];
            uint_32         misc[32];
        } decoded;
    } cr;

    /* -------------------------------------------------------------------- */
    /*  STIC bitmap, 160 x 200.  Also, pointer to graphics object.          */
    /* -------------------------------------------------------------------- */
    uint_8      *disp;
    gfx_t       *gfx;

    /* -------------------------------------------------------------------- */
    /*  Private GROM, GRAM, BACKTAB memory pointers for efficiency.         */
    /* -------------------------------------------------------------------- */
    uint_16     *grom;
    uint_16     *gram;
    uint_16     *backtab;

    /* -------------------------------------------------------------------- */
    /*  IRQ generation (happens every video display update).                */
    /* -------------------------------------------------------------------- */
    int         *irq;

    /* -------------------------------------------------------------------- */
    /*  STIC internal flags.                                                */
    /* -------------------------------------------------------------------- */
    int         phase;          /* 0 == Start of VBlank, 1 == Start of Pic  */
    int         ve_post;        /* Where vid-enables are posted.            */
    int         vid_enable;     /* Must be set every vsync to enable vid.   */
    int         mode;           /* 1 == FG/BG mode, 0 == Color Stack.       */
    int         bt_dirty;       /* BACKTAB is dirty.                        */
    int         gr_dirty;       /* GRAM is dirty.                           */
    int         ob_dirty;       /* MOBs are dirty.                          */
    void        (*upd)(struct stic_t*); /* Update fxn for curr disp mode.   */
} stic_t;

/*
 * ============================================================================
 *  NONE OF THIS TIMING CRAP IS USED.  I'M DOING THIS A MUCH SIMPLER WAY NOW.
 * ============================================================================
 *  The STIC allows timely GRAM servicing only during VBLANK.  Outside 
 *  VBLANK, the CPU is stalled with wait-states.  The following macros
 *  define the time period for VBLANK and DISPACTV.  
 *
 *  TODO:  Consider making these config variables instead.
 *
 *  STIC_COLBURST    -- Color burst frequency (4x the system clock rate)
 *  STIC_FRAMCLKS    -- Number of system clocks in one frame
 *  STIC_DISPACTV    -- Length of Active Display period
 *  STIC_VBLANK      -- Length of Vertical Blanking period.
 *
 *  These values are calculated from the following facts and assumptions.
 *
 *   -- A full NTSC frame has 525 lines divided into two fields,
 *
 *   -- As a result each field has 262.5 lines.
 *
 *   -- The STIC has a vertical resolution of 192 pixels in the graphical
 *      display area.
 *
 *   -- I think we can safely assume that the STIC treats non-graphics card
 *      areas as "blanking" time, even though it is actively outputting the
 *      background color.  (I may end up getting proven wrong here, but the
 *      spec sheet seems to indicate this.)
 *
 *   -- One display period in system clocks is the colorburst rate
 *      divided by 240 for 60Hz NTSC.
 *
 *   -- So, the DISPACTV time is 192/262.5 of the total display period
 *      and VBLANK is the remainder.
 *
 *  Note:  This doesn't really take into account the fact that the active
 *  display period shifts slightly when the vertical delay register is
 *  set (for vertical scrolling).  For the existing set of Intellivision
 *  games, I don't think it really matters.  The _really_ tricky programmers
 *  that relied on exact synchronization behavior between the display chip
 *  and the CPU didn't show up until the C64 demo competitions really took
 *  off, I think.
 * ============================================================================
 */

#define STIC_COLBURST   3579545 /* Exact */
#define STIC_DISPACTV   (((STIC_COLBURST * 384) + 63000) / 126000)
#define STIC_FRAMCLKS   ( (STIC_COLBURST + 120) / 240)
#define STIC_VBLANK     (STIC_FRAMCLKS - STIC_DISPACTV)

/*
 * ============================================================================
 *  STIC_CTRL_RD     -- Read from a STIC control register.
 *  STIC_CTRL_WR     -- Write to a STIC control register.
 *  STIC_BTAB_WR     -- Monitor writes to BACKTAB
 *  STIC_GRAM_WR     -- Monitor writes to GRAM
 * ============================================================================
 */

uint_32 stic_ctrl_rd(periph_t *per, periph_t *ign, uint_32 addr, uint_32 data);
void    stic_ctrl_wr(periph_t *per, periph_t *ign, uint_32 addr, uint_32 data);
void    stic_btab_wr(periph_t *per, periph_t *ign, uint_32 addr, uint_32 data);
void    stic_gram_wr(periph_t *per, periph_t *ign, uint_32 addr, uint_32 data);

/*
 * ============================================================================
 *  STIC_INIT        -- Initialize the STIC
 *  STIC_TICK        -- Perform a STIC update
 * ============================================================================
 */

int stic_init
(
    stic_t          *stic,
    mem_t           *grom,
    mem_t           *gram,
    int             *irq,
    gfx_t           *gfx
);

sint_32 stic_tick
(
    periph_p        per,
    sint_32         len
);

#endif

/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */
